#include "lookup_notif.h"
#include "symbian_auto_ptr.h"
#include "blackboardclientsession.h"
#include "contextcommon.h"
#include "blackboard_cs.h"
#include <s32mem.h>
#include "bbutil.h"
#include "break.h"
#include "bbtypes.h"
#include "csd_presence.h"

class CLookupHandlerImpl : public CLookupHandler {
	CLookupHandlerImpl(MOutgoingLookup& aOutgoing);
	void ConstructL();
	~CLookupHandlerImpl();

	virtual void IncomingLookupL(const MBBData* aData);
	virtual void GetNextLookupL(); // will do an async callback with the next one
	virtual void AckLookupL(TUint aId);

	void CheckedRunL();
	void ConnectL();
	void SetFilterL();
	void WaitForNotify();
	void WriteL(const TDesC& aBuf);
	void DoCancel();

	RBBClient	iBBClient;

	HBufC8*		iSerializedData;
	TPtr8		iP;

	TFullArgs	iFullArgs;
	TInt		iAsyncErrorCount;
	MBBDataFactory*	iFactory;
	MOutgoingLookup& iOutgoing;
	MBBData		*iWaitingData; TInt iWaitingId;

	friend CLookupHandler;
	friend auto_ptr<CLookupHandlerImpl>;
};

_LIT(KLookupHandler, "CLookupHandler");

CLookupHandler::CLookupHandler() : CCheckedActive(EPriorityStandard, KLookupHandler) { }

CLookupHandler* CLookupHandler::NewL(MOutgoingLookup& aOutgoing)
{
	auto_ptr<CLookupHandlerImpl> ret(new (ELeave) CLookupHandlerImpl(aOutgoing));
	ret->ConstructL();
	return ret.release();
}


CLookupHandlerImpl::CLookupHandlerImpl(MOutgoingLookup& aOutgoing) : 
	iP(0, 0), iOutgoing(aOutgoing) { }

void CLookupHandlerImpl::ConstructL()
{
	iSerializedData=HBufC8::NewL(2048);

	iFactory=CBBDataFactory::NewL();

	CActiveScheduler::Add(this);
	ConnectL();
	SetFilterL();
}

CLookupHandlerImpl::~CLookupHandlerImpl()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("~CLookupHandlerImpl"));

	Cancel();
	iBBClient.Close();
	delete iSerializedData;
	delete iFactory;
}

void CLookupHandlerImpl::IncomingLookupL(const MBBData* aData)
{
	if (!aData) return;

	auto_ptr<HBufC8> serialized(HBufC8::NewL(1024));
	TInt err=KErrOverflow;
	while (err==KErrOverflow) {
		TPtr8 bufp(serialized->Des());
		RDesWriteStream ws(bufp);

		aData->Type().ExternalizeL(ws);
		CC_TRAP(err, aData->ExternalizeL(ws));
		if (err==KErrNone) {
			ws.CommitL();
		} else if (err==KErrOverflow) {
			serialized->Des().Zero();
			serialized.reset( serialized->ReAllocL(
				serialized->Des().MaxLength()*2) );
		}
	}
	User::LeaveIfError(err);

	TUint id;
	TRequestStatus s;

	TBool replace=EFalse;
	TTime expires; expires.HomeTime(); expires+=TTimeIntervalDays(7);

	iBBClient.Put(KIncomingLookupTuple, KNullDesC,
		KNoComponent, *serialized, EBBPriorityNormal,
		replace, id, s, expires, ETrue, EFalse);
	User::WaitForRequest(s);
	User::LeaveIfError(s.Int());
}

void CLookupHandlerImpl::GetNextLookupL()
{
	if (iWaitingData) {
		TRequestStatus* s=&iStatus;
		User::RequestComplete(s, KErrNone);
		SetActive();
	} else {
		WaitForNotify();
	}
}

void CLookupHandlerImpl::AckLookupL(TUint aId)
{
	if (aId==0) return;

	if (aId==iWaitingId) {
		delete iWaitingData; iWaitingData=0;
		iWaitingId=0;
	}

	TRequestStatus s;
	iBBClient.Delete(aId, s);
	User::WaitForRequest(s);
	if (s==KErrNotFound) return;
	User::LeaveIfError(s.Int());
}

void CLookupHandlerImpl::ConnectL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("ConnectL"));

	Cancel();
	TInt errorcount=0;
	TInt err=KErrNone;
	while (errorcount<5) {
		iBBClient.Close();
		err=iBBClient.Connect();
		if (err==KErrNone) return;
		errorcount++;
	}
	User::Leave(err);
}

void CLookupHandlerImpl::SetFilterL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("SetFilterL"));

	TInt errorcount=0, err=KErrNone;
	while (errorcount<5) {
		TRequestStatus s;
		iBBClient.AddNotificationL(KOutgoingLookupTuple, ETrue, EBBPriorityNormal, s);
		User::WaitForRequest(s);
		err=s.Int();
		if (err==KErrNone) return;
		ConnectL();
		errorcount++;
	}
	User::Leave(err);
}

void CLookupHandlerImpl::WaitForNotify()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("WaitForNotify"));

	if (IsActive()) return;

	iSerializedData->Des().Zero();
	iP.Set(iSerializedData->Des());
	iBBClient.WaitForNotify(iFullArgs, iP, iStatus);
	SetActive();
}

void CLookupHandlerImpl::DoCancel()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("DoCancel"));

	iBBClient.CancelNotify();
}


_LIT(KValue, "value");

void CLookupHandlerImpl::CheckedRunL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("CheckedRunL"));

	if (iWaitingData) {
		TInt err;
		CC_TRAP(err, iOutgoing.OutgoingLookupL(iWaitingId, iWaitingData));
		return;
	}
	{
		if (iStatus.Int()!=KErrNone) {
			if (iStatus.Int()==EBufferTooSmall) {
				iSerializedData->Des().Zero();
				iSerializedData=iSerializedData->ReAllocL(iSerializedData->Des().MaxLength()*2);
				iAsyncErrorCount=0;
				WaitForNotify();
				return;
			}
			if (iAsyncErrorCount>5) User::Leave(iStatus.Int());
			ConnectL();
			SetFilterL();
			WaitForNotify();
			return;
		}
	}

	MBBData* d=0;
	{
		RDesReadStream rs(*iSerializedData);
		CleanupClosePushL(rs);
		TTypeName read_type=TTypeName::IdFromStreamL(rs);
		if (! (read_type==KPresenceType) ) {
			{
				d=iFactory->CreateBBDataL(read_type, KValue, iFactory);
				CleanupPushBBDataL(d);
			}
			{
				d->InternalizeL(rs);
			}
			CleanupStack::Pop();
		} else {
			AckLookupL(iFullArgs.iId);
			WaitForNotify();
			CleanupStack::PopAndDestroy();
			return;
		}
		CleanupStack::PopAndDestroy();
	}

	{
		TInt err;
		bb_auto_ptr<MBBData> p(d);
		CC_TRAP(err, iOutgoing.OutgoingLookupL(iFullArgs.iId, d));
		if (err!=KErrNone) {
			iWaitingData=p.release();
			iWaitingId=iFullArgs.iId;
		}
	}

}
